'use strict';

// ** Dependencies
const _ = require('lodash');
const util = require('util');
const functions = require('./functions');
const errors = require('./errors');
const Promise = require('bluebird');
const logger = require('./logging').createLogger('activities');
const EventEmitter = require('eventemitter2');


/**
 * Inspect the function signature to build a list of parameters
 * @param func
 */
function getParameters(func) {

    const parameterNames = functions.getParameterNames(func);
    logger.debug('PARAMETER_NAMES', parameterNames);

    const ret = {};

    _.forEach(parameterNames, name => {
        ret[name] = {
            defaultValue: undefined,
            required: false
        };
    });

    return ret;
}

function setParameters(parameters, args) {

    const ret = {};

    // Set the value of each parameter from the args, or use defaults
    _.forEach(parameters, (param, name) => {
        logger.debug('PARAM', param);

        // Use value from args or apply a default value
        const value = args.hasOwnProperty(name)
            ? args[name]
            : param.defaultValue;

        ret[param] = value;
    });

    return ret;
}

function buildContext(context, params, options) {

    if (util.isNullOrUndefined(context))
        context = {};

    context['__params'] = params;
    context['__options'] = options;

    return context;
}

class Activity extends EventEmitter {
    constructor(name, handler, parameters) {
        super();

        this._name = name;
        this._handler = handler;
        this._parameters = parameters;

        // Set run function so we can capture 'this' from the caller
        const self = this;
        this.run = function (args, options) {

            // Default options (for the run command)
            options = options || {};

            return new Promise((resolve, reject) => {

                try {
                    // Build an array of arguments to pass to the caller
                    const params = setParameters(self._parameters, args);
                    logger.debug('PARAMS', params);

                    // Create context for the activity
                    const context = buildContext(this, params, options);
                    logger.debug('CONTEXT', context);

                    // Convert parameters to array to apply to function
                    const argList = _.values(params);
                    logger.debug('ARG_LIST', argList);

                    // Send 'run' notification event
                    self.emit('run', {
                        params: params,
                        options: options
                    });

                    // Invoke the function with the specified context
                    const result = self._handler.apply(context, argList);

                    // Send 'result' event notification
                    self.emit('result', result);

                    resolve(result);
                } catch (err) {
                    // Send 'error' event notification
                    self.emit('error', errors('TASK_ERROR', 'An error occurred running the task.', {
                        name: self.name,
                        error: err
                    }));

                    reject(err);
                }
            });
        };
    }

    get name() {
        return this._name;
    }
}

function isActivity(obj) {
    return obj instanceof Activity;
}

function createActivity(name, handler, options) {

    // createCommand(function) -> createCommand(function.name, function)
    if (arguments.length === 1 && util.isFunction(arguments[0]))
        return createActivity(arguments[0].name, arguments[0]);

    // createCommand({name:, handler:, ...} -> createCommand(.name, .handler, {...})
    if (arguments.length === 1 && util.isObject(arguments[0]))
        return createActivity(arguments[0].name, arguments[0].handler, arguments[0]);

    // Argument Defaults
    options = options || {};

    // Name is a required argument
    if (_.isEmpty(name))
        throw errors('ARGUMENT_REQUIRED', 'The argument "name" is required.', {name: name});

    // Handler must be a function
    if (!util.isFunction(handler))
        throw errors('ARGUMENT_INVALID', 'The argument "handler" must be a function.', {handler: handler});

    // Determine the parameters for the activity
    const parameters = options.parameters || getParameters(handler);
    logger.debug('PARAMETERS', parameters);

    const activity = new Activity(name, handler, parameters);
    return activity;
}

// ** Exports
module.exports = createActivity;
module.exports.isActivity = isActivity;
module.exports.create = createActivity;